<?php
class WEAPP{
	use WPJAM_Instance_Trait;

	private $appid;
	private $secret;
	private $api_host	= 'https://api.weixin.qq.com';

	public function __construct($appid, $secret=''){
		$this->appid	= $appid;
		$this->secret	= $secret;
	}

	public function __call($method, $args){
		if(in_array($method, ['cache_get', 'cache_set', 'cache_add', 'cache_delete'])){
			$object	= WPJAM_Cache_Group::get_instance('weapp', [
				'global'	=> true, 
				'prefix'	=> $this->appid
			]);

			return call_user_func_array([$object, $method], $args);
		}elseif(in_array($method, ['decrypt_user', 'decrypt_gid', 'decrypt_run', 'decrypt_phone'])){
			return call_user_func_array([$this, 'decrypt'], $args);
		}
	}

	public function get_component_blog_id(){
		return weapp_get_setting('component_blog_id', $this->appid);
	}

	public function jscode2session($code){
		$session	= $this->cache_get('session:'.$code);

		if($session === false || empty($session['openid']) || empty($session['session_key'])){
			$component_blog_id = $this->get_component_blog_id();

			if($component_blog_id){
				$session	= apply_filters('weapp_session_data', null, $code, $this->appid, $component_blog_id);
			}else{
				$session	= $this->http_request('/sns/jscode2session?js_code='.$code.'&grant_type=authorization_code', ['using_secret'=>true]);
			}

			if(is_wp_error($session)){
				return $session;
			}

			if(empty($session['openid']) || empty($session['session_key'])){
				trigger_error(var_export($session, true));
				return new WP_Error('invalid_session_data', '无效的 session_data');
			}

			$this->cache_set('session:'.$code, $session, MINUTE_IN_SECONDS);
			$this->cache_set('session:'.$session['openid'], $session, DAY_IN_SECONDS);
		}

		return $session;
	}

	public function get_session_by_openid($openid){
		$session	= $this->cache_get('session:'.$openid);

		if($session === false){
			return new WP_Error('empty_session_key', '服务器缓存的 Session Key 失效！');
		}

		return $session;
	}

	public function decrypt($aeskey, $iv, $encrypted){
		if(strlen($aeskey) != 24){
			return new WP_Error('illegal_aes_key', '非法 AES Key');
		}

		if(strlen($iv) != 24){
			return new WP_Error('illegal_iv', '非法 iv');
		}

		$aeskey	= base64_decode($aeskey);
		$iv		= base64_decode($iv);
		// $encrypted	= base64_decode($encrypted);

		try{
			$decrypted	= openssl_decrypt($encrypted, "AES-128-CBC", $aeskey, 0, $iv);
		}catch(Exception $e) {
			return new WP_Error('aes_decrypt_failed', 'AES 解密失败');
		}

		if(!$decrypted){
			return new WP_Error('aes_decrypt_failed', 'AES 解密失败');
		}

		if (strlen($decrypted) < 16)	{
			return new WP_Error('illegal_decrypted_length', '非法解密数据长度');
		}

		$decrypted	= wpjam_json_decode($decrypted);

		if(is_wp_error($decrypted)){
			return $decrypted;
		}

		if($decrypted['watermark']['appid'] != $this->appid){
			return new WP_Error('illegal_watermark_appid', '和数据水印中的 appid 不匹配！');
		}

		return $decrypted;
	}

	public function get_access_token($force=false, $return='token'){
		$response	= weapp_get_setting('access_token', $this->appid);

		if($force || !$response || !is_array($response) || !isset($response['expires_at']) || $response['expires_at'] <= time()){
			$component_blog_id	= $this->get_component_blog_id();

			if($component_blog_id){
				$response	= apply_filters('weapp_access_token', null, $this->appid, $component_blog_id);
			}else{
				$response	= $this->http_request('/cgi-bin/token?grant_type=client_credential', ['using_secret'=>true]);
			}

			if(is_wp_error($response)){
				return $response;
			}

			$response['expires_at']	= $response['expires_in']-60+time();

			if(!$force){
				weapp_update_setting('access_token', $response, $this->appid);
			}
		}

		return $return == 'array' ? $response : $response['access_token'];
	}

	public function stream($path, $data, $method='POST'){
		$type		= array_pull($data, 'type');
		$media_id	= $data['media_id'] ?? md5(maybe_serialize($data));
		$file		= $this->get_media_file($media_id, $type);

		if(!file_exists($file) || $type == 'unlimit'){
			if(isset($data['time'])){
				unset($data['time']);
			}

			$result	= $this->http_request($path, [
				'method'	=> $method,
				'body'		=> $data,
				'stream'	=> true,
				'filename'	=> $file,
			]);

			if(is_wp_error($result)){
				if(file_exists($file)){
					unlink($file);
				}
				
				return $result;
			}
		}

		return $media_id;
	}

	public function post($path, $data=[], &$headers=[]){
		return $this->http_request($path, ['method'=>'POST', 'body'=>$data], $headers);
	}

	public function get($path, $args=[], &$headers=[]){
		$args	= array_filter($args);
		$path	= $args ? add_query_arg($args, $path) : $path;

		return $this->http_request($path, [], $headers);
	}

	public function file($path, $media, $name='media', $timeout=5){
		return $this->http_request($path, [
			'method'	=> 'FILE',
			'timeout'	=> $timeout,
			'body'		=> [$name => new CURLFile($media, '', basename($media))]
		]);
	}

	public function http_request($path, $args=[], &$headers=[]){
		$args	= wp_parse_args($args, [
			'need_json_encode'	=> true,
			'timeout'			=> 5
		]);

		if(array_pull($args, 'using_secret')){
			$path	= add_query_arg(['appid'=>$this->appid, 'secret'=>$this->secret], $path);
		}else{
			$access_token = $this->get_access_token();

			if(is_wp_error($access_token)){
				return $access_token;
			}

			$path	= add_query_arg(['access_token'=>$access_token], $path);
			$path	= str_replace('%40', '@', $path);  
		}

		if(!str_starts_with($path, 'http')){
			$path	= $this->api_host.$path;
		}

		$response	= wpjam_http_request($path, $args, [], $headers);

		if(is_wp_error($response)){
			$errcode	= $response->get_error_code();

			if($errcode == '40001' || $errcode == '40014' || $errcode == '42001'){
				// 40001 获取access_token时AppSecret错误，或者access_token无效。请开发者认真比对AppSecret的正确性，或查看是否正在为恰当的公众号调用接口
				// 40014 不合法的access_token，请开发者认真比对access_token的有效性（如是否过期），或查看是否正在为恰当的公众号调用接口
				// 42001 access_token超时，请检查access_token的有效期，请参考基础支持-获取access_token中，对access_token的详细机制说明
				weapp_delete_setting('access_token', $this->appid);
			}elseif($errcode == '50002' || $errcode == '9009205' || $errcode = '9009301'){
				// 50002 用户受限，可能是违规后接口被封禁
			}

			// trigger_error(var_export($response, true));
		}

		return $response;
	}

	public function get_media_dir($file_name, $type=''){
		$base_dir	= WEAPP_MEDIA_DIR.'/'.$this->appid;

		if(!is_dir($base_dir)){
			mkdir($base_dir, 0777, true);
		}

		$media_dir	= '';

		if($type){
			$media_dir	.= '/'.$type;

			if(!is_dir($base_dir.$media_dir)){
				mkdir($base_dir.$media_dir, 0777, true);
			}
		}

		$media_dir	.= '/'.substr($file_name, 0, 2);

		if(!is_dir($base_dir.$media_dir)){
			mkdir($base_dir.$media_dir, 0777, true);
		}

		return $media_dir;
	}

	public function get_media_url($media_id, $type=''){
		return WEAPP_MEDIA_URL.'/'.$this->appid.$this->get_media_dir($media_id, $type).'/'.$media_id.'.jpg';
	}

	public function get_media_file($media_id, $type=''){
		return WEAPP_MEDIA_DIR.'/'.$this->appid.$this->get_media_dir($media_id, $type).'/'.$media_id.'.jpg';
	}

	public static function get_instance($appid, $secret){
		$instance	= self::instance_exists($appid);

		return $instance ?: self::add_instance($appid, new self($appid, $secret));
	}
}

class WEAPP_Setting extends WPJAM_Model{
	const SOURCE_DEFAULT	= 0;
	const SOURCE_AUTHORIZE	= 1;

	public static function insert($data){
		$data['time']	= time();

		if(empty($data['blog_id'])){
			$data['blog_id']	= get_current_blog_id();
		}

		if(self::get_by('blog_id', $data['blog_id'])){
			return new WP_Error('weapp_exists', '该站点已有小程序。');
		}

		if(empty($data['component_blog_id']) && !weapp_exists($data['appid'], $data['secret'])){
			return new WP_Error('weapp_not_exists', '输入的 appid 和 secret 有误，请仔细核对！');
		}

		return parent::insert($data);
	}

	public static function update($appid, $data){
		$setting	= self::get($appid);

		if(!$setting){
			return new WP_Error('weapp_setting_not_exists', '系统中没有你更新的小程序，可能已经被删除了。');
		}

		if(empty($setting['component_blog_id']) && empty($data['component_blog_id']) && isset($data['secret'])){
			if($data['secret'] != $setting['secret'] && !weapp_exists($appid, $data['secret'])){
				return new WP_Error('weapp_not_exists', '输入的 appid 和 secret 有误，请仔细核对！');
			}
		}

		return parent::update($appid, $data);
	}

	public static function get_handler(){
		return wpjam_get_handler([
			'table_name'		=> self::get_table(),
			'primary_key'		=> 'appid',
			'lazyload_key'		=> ['blog_id'=>'site'],
			'cache_key'			=> 'blog_id',
			'orderby'			=> 'time',
			'cache_group'		=> ['weapp_settings', true],
			'field_types'		=> ['blog_id'=>'%d','time'=>'%d'],
			'searchable_fields'	=> ['appid','name'],
			'filterable_fields'	=> ['component_blog_id'],
		]);
	}

	public static function get_table(){
		return $GLOBALS['wpdb']->base_prefix . 'weapps';
	}

	public static function create_table($appid=''){
		require_once(ABSPATH . 'wp-admin/includes/upgrade.php');

		$table	= self::get_table();

		if($GLOBALS['wpdb']->get_var("show tables like '{$table}'") != $table) {
			$sql = "
			CREATE TABLE IF NOT EXISTS `{$table}` (
				`blog_id` bigint(20) NOT NULL,
				`name` varchar(255) NOT NULL,
				`appid` varchar(32) NOT NULL,
				`secret` varchar(40) NOT NULL,
				`component_blog_id` bigint(20) NOT NULL DEFAULT 0,
				`time` int(10) NOT NULL,

				PRIMARY KEY	(`appid`),
				KEY `type` (`type`),
				KEY `blog_id` (`blog_id`)
			) DEFAULT CHARSET=utf8mb4 COLLATE=utf8mb4_unicode_ci;
			";
	 
			dbDelta($sql);
		}
	}

	public static function render_item($item){
		$item['time']	= get_date_from_gmt(date('Y-m-d H:i:s',$item['time']));

		if($detail = get_blog_details($item['blog_id'])){
			$item['name']		= $item['name'] ?: $detail->blogname;
			$item['name']		= '<a href="'.get_admin_url($item['blog_id'],'page=weapp-settings').'">'.$item['name'].'</a>';
			$item['blog_id']	= $detail->blogname;
		}else{
			$item['name']		= '<span class="red">站点已经删除</span>';
		}

		if($item['component_blog_id']){
			unset($item['row_actions']['edit']);
			unset($item['row_actions']['delete']);
		}

		return $item;
	}

	public static function get_actions(){
		$actions	= [
			'add'		=> ['title'=>'新增'],
			'edit'		=> ['title'=>'编辑'],
			'delete'	=> ['title'=>'删除',	'direct'=>true,	'confirm'=>true],
		];

		if(self::get_by('blog_id', get_current_blog_id())){
			unset($actions['add']);
		}

		return $actions;
	}

	public static function get_fields($action_key='', $id=''){
		return [
			'name'				=> ['title'=>'小程序名',		'type'=>'text',	'show_admin_column'=>true,	'class'=>'all-options',	'required'],
			'appid'				=> ['title'=>'小程序ID',		'type'=>'text',	'show_admin_column'=>true,	'class'=>'all-options',	'required'],
			'secret'			=> ['title'=>'小程序密钥',	'type'=>'text',	'required'],
			'blog_id'			=> ['title'=>'所属站点',		'type'=>'text',	'show_admin_column'=>'only'],
			'component_blog_id'	=> ['title'=>'第三方平台',	'type'=>'text',	'show_admin_column'=>'only'],
			'time'				=> ['title'=>'添加时间',		'type'=>'view',	'show_admin_column'=>'only',	'sortable_column'=>true],
		];
	}

	public static function get_list_table(){
		return [
			'title'			=> '小程序',
			'singular'		=> 'weapp-setting',
			'plural'		=> 'weapp-settings',
			'primary_key'	=> 'appid',
			'model'			=> self::class,
			'capability'	=> 'manage_sites',
		];
	}

	public static function get_settings($blog_id){	// 兼容
		return weapp_get_settings($blog_id);
	}
}
